#include "defs.h"
static chiffre un[] = {1};

                     /* +---------------------+
                        |  Division de Knuth  |
                        +---------------------+ */

/* q <- trunc(a/b), a <- a - q*b, lg(a) = 2n, lg(b) = lg(q) = n */
/* le chiffre de tte de b doit avoir son bit de poids fort  1 */
void div_knuth(chiffre *q, chiffre *a, chiffre *b, longueur n) {
  longueur i,j;
  ndouble r,s,x;
  zdouble z;

  /* division 2 chiffres par 1 chiffre */
  if (n <= 1) {
    if (aa(1) >= bb(0)) {
      q[0] = mask;
      r = aa(0)+bb(0) + ((aa(1)-bb(0)) >> shift);
      a[0] = r;
      a[1] = r >> shift;
    }
    else {
      r = (aa(1) << shift) + aa(0);
      q[0] = r/bb(0);
      a[1] = 0;
      a[0] = r%bb(0);
    }
  }

  /* division  plusieurs chiffres */
  else {
    for (i=n-1; i>=0; i--) {

      /* divise les trois chiffres de tte de a par les deux chiffres de b */
      if (aa(n+i) >= bb(n-1)) x = mask;
      else {
	r = (aa(n+i) << shift) + aa(n+i-1);
	x = r/bb(n-1);
	r = ((r%bb(n-1)) << shift) + aa(n+i-2);
	s = x*bb(n-2);
	while (r < s) {x--; s -= r; r = (bb(n-1)<<shift) + bb(n-2);}
      }
      
      /* a <- a - x*base^i */
      for (j=0,r=0; j<n; j++) {
	r += x*bb(j);
	z = aa(i+j) - (r&mask);
	a[i+j] = z;
	r >>= shift; if (z<0) r++;
      }
      if (r != aa(n+i)) {inc(a+i,b,n,n); x--;} /* si ngatif ajoute b*base^i */
      a[n+i] = 0;
      
      q[i] = x;
    }
  }
}

       /* +-------------------------------------------------+
          |  Multiplication disymtrique, lb-1 <= la <= lb  |
          +-------------------------------------------------+ */

static inline
void multiply(chiffre *c, chiffre *a, chiffre *b, longueur la, longueur lb, longueur n0) {

  if (la == lb) karatsuba(c,a,b,la,n0);
  else {
    chiffre *a1 = getmem(lb);
    movemem(a1,a,la); a1[la] = 0;
    karatsuba(c,a1,b,lb,n0);
    freemem(a1);
  }
}

              /* +-----------------------------------+
                 |  Division de Burnikel et Ziegler  |
                 +-----------------------------------+ */


/* c <- trunc(a/b), lg(a) = 2n, lg(b) = lg(q) = n                 */
/* si rem = 0 alors a <- a - high(b*c) sinon a <- a - b*c         */
/* si rem = 0 le quotient est trop grand d'au plus 2log(n) units */
/* le chiffre de tte de b doit avoir son bit de poids fort  1   */
void burnikel(int rem, chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0) {

  if (n <= n0) div_knuth(c,a,b,n);
  else {
    longueur q=n/2, p=n-q, i,j;
    chiffre *buff = getmem(n+1);
    zdouble r;

                                          /* premier chiffre        */
    if (compare(a+n+q,b+q,p) >= 0) {      /* blocs de tte gaux ?  */
      clearmem(c+q,p);                    /* => tronque le quotient */
      r = dec(a+n,b,n,n);
      while(r) {dec(c+q,un,p,1); r += inc(a+q,b,n+p,n);}
    }
    else {                                /* sinon,                 */
      burnikel(1, c+q, a+2*q, b+q, p,n0); /* divise les poids forts */
      multiply(buff, b, c+q, q,p,n0);     /* retranche b0*c1        */
      r = dec(a+q, buff, n,n);
      while (r) {dec(c+q, un, p,1); r += inc(a+q, b, n,n);}
    }
                                          /* deuxime chiffre       */
    if (compare(a+n,b+p,q) >= 0) {        /* blocs de tte gaux ?  */
      for (i=0; i<q; i++) c[i] = mask;    /* => tronque le quotient */
      clearmem(a+n,q);
      for (i=q; i>0;) {
	j = ((i <= n0) || (rem)) ? i : (i+1)/2;
	inc(a+n-j,b+n-i,j+1,j);
	i -= j;
      }
    }
    else burnikel(rem, c, a+p, b+p, q, n0); /* sinon divise */

    if (rem) {
      multiply(buff, c,b, q,p,n0);         /* retranche b0*c0  */
      r = dec(a, buff, n+1,n);
      while (r) {dec(c, un, n,1); r += inc(a, b, n+1,n);}
    }

    freemem(buff);
  }
}

/* corrige la division sans reste */
void burnikel_correct(chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0) {

  if (n > n0) {
    longueur q=n/2, p=n-q;
    chiffre *buff;
    zdouble r;

    burnikel_correct(c,a+p,b+p,q,n0);

    buff = getmem(n+1);
    multiply(buff, c,b, q,p,n0);
    r = dec(a, buff, n+1,n);
    while (r) {dec(c, un, n,1); r += inc(a, b, n+1,n);}
    freemem(buff);
  }
}

                    /* +-----------------------+
                       |  Division sans reste  |
                       +-----------------------+ */

void burnikel_nr(chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0) {

  if (n <= n0) div_knuth(c,a,b,n);
  else {

    /* effectue la division sans reste avec deux chiffres aprs la virgule */
    chiffre *a1 = getmem(2*n+4);
    chiffre *b1 = getmem(n+2);
    chiffre *c1 = getmem(n+2);

    movemem(a1+4,a,2*n); clearmem(a1,4);
    movemem(b1+2,b,n);   clearmem(b1,2);
    burnikel(0,c1,a1,b1,n+2,n0);

    /* si la partie dcimale est trop petite corrige la division */
    if ((c1[1] & mask) == 0) burnikel_correct(c1,a1,b1,n+2,n0);

    /* tronque le rsultat */
    movemem(c,c1+2,n);

    freemem(a1); freemem(b1); freemem(c1);
  }
}

/* test */
double test_burnikel(int rem, longueur n, longueur n0, int nrep, int check) {
  chiffre *a,*b,*q,*r;
  double t;

  b = random_bits(n);   b[n-1] = bb(n-1) | msb;
  if (rem < 0) {
    q = random_bits(n);
    a = getmem(2*n);
    karatsuba(a,q,b,n,n0);
    rem = 0;
  }
  else {
    a = random_bits(2*n); a[2*n-1] = aa(2*n-1) % bb(n-1);
    q = getmem(n);
  }
  r = getmem(2*n);

  if (check) {
    chiffre *x = getmem(2*n);
    int ok;

    movemem(r,a,2*n);

    if (rem) {
      burnikel(1,q,r,b,n,n0);
      ok = (sub(x,b,r,n,n) == 0);
      if (ok) {
	muln2(x,q,b,n);
	inc(x,r,2*n,2*n); 
	ok = (compare(a,x,2*n) == 0);
      }
    }

    else {
      burnikel_nr(q,r,b,n,n0);
      movemem(r,a,2*n);
      div_knuth(x,r,b,n);
      ok = (compare(q,x,n) == 0);
    }

    if (!ok) {
      fprintf(stderr,"division fausse\n");
      fprintf(stderr,"a="); dump(a,2*n);
      fprintf(stderr,"b="); dump(b,n);
      fprintf(stderr,"q="); dump(q,n);
      fprintf(stderr,"r="); dump(r,2*n);
      fflush(stderr);
      exit(1);
    }
    
    t = 0;
    freemem(x);
  }

  else {
    int i;
    chrono();
    for (i=0; i < nrep; i++) {
      movemem(r,a,2*n);
      if (rem) burnikel(1,q,r,b,n,n0); else burnikel_nr(q,r,b,n,n0);
    }
    t = chrono()/nrep;
  }

  freemem(a); freemem(b); freemem(q); freemem(r);
  return(t);
}
