                   /* +------------------------+
                      |  Dfinitions communes  |
                      +------------------------+ */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#ifdef bits64
/* machine 64 bits -> codage sur 32 bits (non test) */
typedef unsigned int chiffre;
typedef unsigned long ndouble;
typedef signed   long zdouble;
typedef signed   long longueur;
#define nbits  64
#define mask   0xffffffff
#define msb    0x80000000
#define shift  32

#else
/* machine 32 bits -> codage sur 16 bits */
typedef unsigned int chiffre;
typedef unsigned long ndouble;
typedef signed   long zdouble;
typedef signed   long longueur;
#define nbits  32
#define mask   0xffff
#define msb    0x8000
#define shift  16
#endif

/* mon compilateur C (egcs-2.91.66) est bogu : 
   pour a de type chiffre *, a[0] -> nb sur 16 bits
                             a[i] -> nb sur 32 bits...
  contre-mesure : charger explicitement 16 bits. */

#define aa(i) (a[i] & mask)
#define bb(i) (b[i] & mask)
#define cc(i) (c[i] & mask)
#define dd(i) (d[i] & mask)

                          /* +-----------+
                             |  utils.c  |
                             +-----------+ */

double  chrono();                          /* temps pass en milisecondes   */
void    dump(chiffre *a, longueur n);      /* affichage hexadcimal         */
void    bloc_dump(chiffre *a, longueur n, longueur n0);
chiffre *from_string(char *s, longueur n); /* chane hexadcimale -> nombre */
chiffre *random_bits(longueur n);          /* initialisation alatoire      */
int     compare(chiffre *a, chiffre *b, longueur n);       /* compare a et b */

                         /* +------------+
                            |  memory.c  |
                            +------------+ */

chiffre *getmem(longueur n);              /* alloue n chiffres */
#define freemem(x) free(x)
void    clearmem(chiffre *a, longueur n); /* mise  zro       */
void    movemem(chiffre *a, chiffre *b, longueur n); /* a <- b */

                          /* +-----------+
                             |  blocs.c  |
                             +-----------+ */

/* dcompose a en blocs de taille n0 chiffres + 1 chiffre de garde */
void expand(chiffre *b, chiffre *a, longueur la, longueur lb, longueur n0);

/* regroupe les blocs de taille n0 avec un dcalage n1 */
void compact(chiffre *b, chiffre *a,
	     longueur la,  /* nb de blocs       */
	     longueur n0,  /* taille des blocs  */
	     longueur n1,  /* dcalage          */
	     longueur lb); /* longueur rsultat */

                        /* +--------------+
                           |  addition.c  |
                           +--------------+ */

/* c <- a op b, retourne la retenue */
chiffre add(chiffre *c, chiffre *a, chiffre *b, longueur la, longueur lb);
zdouble sub(chiffre *c, chiffre *a, chiffre *b, longueur la, longueur lb);

/* a <- a op b, retourne la retenue */
chiffre inc(chiffre *a, chiffre *b, longueur la, longueur lb);
zdouble dec(chiffre *a, chiffre *b, longueur la, longueur lb);

/* c <- |a-b|, retourne le signe */
int asub(chiffre *c, chiffre *a, chiffre *b, longueur la, longueur lb);

/* oprations par blocs */
void bloc_add(chiffre *c, chiffre *a, chiffre *b,  longueur la, longueur lb, longueur n0);
void bloc_sub(chiffre *c, chiffre *a, chiffre *b,  longueur la, longueur lb, longueur n0);
void bloc_inc(chiffre *a, chiffre *b, longueur la, longueur lb, longueur n0);
void bloc_dec(chiffre *a, chiffre *b, longueur la, longueur lb, longueur n0);

/* oprations mixtes, a compact, b dcompos en blocs positifs de */
/* longueur n0 avec un dcalage de n1.                             */
chiffre mix_inc(chiffre *a, chiffre *b, longueur la, longueur lb, longueur n0, longueur n1);
zdouble mix_dec(chiffre *a, chiffre *b, longueur la, longueur lb, longueur n0, longueur n1);

                          /* +-----------+
                             |  muln2.c  |
                             +-----------+ */

/* c <- ab, lg(a) = lg(b) = n */
void muln2( chiffre *c, chiffre *a, chiffre *b, longueur n);
void muln2h(chiffre *c, chiffre *a, chiffre *b, longueur n); /* p. haute   */
void muln2l(chiffre *c, chiffre *a, chiffre *b, longueur n); /* p. basse   */
void muln2m(chiffre *c, chiffre *a, chiffre *b, longueur n); /* p. mdiane */
void sqrn2(chiffre *c, chiffre *a,              longueur n); /* carr      */

/* multiplication par blocs en complment  2 */
void bloc_muln2(chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0);

                        /* +---------------+
                           |  karatsuba.c  |
                           +---------------+ */

void   karatsuba     (chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0);
void   karasquare    (chiffre *c, chiffre *a,             longueur n, longueur n0);
void   bloc_karatsuba(chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0);
double test_karatsuba     (longueur n, longueur n0, int nrep, int check);
double test_karasquare    (longueur n, longueur n0, int nrep, int check);
double test_bloc_karatsuba(longueur n, longueur n0, int nrep, int check);

                         /* +-------------+
                            |  mulders.c  |
                            +-------------+ */

void   mulders_low(     chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0);
void   mulders_high(    chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0);
void   raw_mulders_low( chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0);
void   bloc_mulders_low(chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0);
double test_mulders(int meth, longueur n, longueur n0, int nrep, int check);

                           /* +---------+
                              |  rmp.c  |
                              +---------+ */

void bloc_rmp1(         chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0);
void bloc_rmp2(         chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0);
void bloc_rmp3(         chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0);
void bloc_rmp4(int big, chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0);
void rmp(     int meth, chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0);
double test_bloc_rmp(int meth, longueur n, longueur n0, int nrep, int check);

                        /* +--------------+
                           |  burnikel.c  |
                           +--------------+ */

void div_knuth(chiffre *q, chiffre *a, chiffre *b, longueur n);
void burnikel(int rem, chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0);
void burnikel_correct( chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0);
void burnikel_nr(      chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0);
double test_burnikel(int rem, longueur n, longueur n0, int nrep, int check);

                         /* +------------+
                            |  rmpdiv.c  |
                            +------------+ */

void   rmpdiv(   chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0);
void   rmpdiv_nr(chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0);
double test_rmpdiv(int rem, longueur n, longueur n0, int nrep, int check);

                       /* +----------------+
                          |  zimmermann.c  |
                          +----------------+ */

void sqrtn2(          chiffre *c, chiffre *a, longueur n);
void sqrtrec(int rem, chiffre *c, chiffre *a, longueur n, longueur n0);
void sqrt_correct(    chiffre *c, chiffre *a, longueur n, longueur n0);
void sqrt_nr(         chiffre *c, chiffre *a, longueur n, longueur n0);
double test_sqrt(int rem, longueur n, longueur n0, int nrep, int check);
