                /* +-------------------------------+
                   |  Multiplication de Karatsuba  |
                   +-------------------------------+ */

#include "defs.h"

/* c <- ab, lg(a) = lg(b) = n */
void karatsuba(chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0) {

  /* cas de base */
  if (n <= n0) {
    longueur i,j;
    ndouble r;

    clearmem(c,n);
    for (i=0; i < n; i++,a++,c++) {
      for (j=0,r=0; j < n; j++) {r+= aa(0)*bb(j) + cc(j); c[j] = r; r >>= shift;}
      c[n] = r;
    }
  }

  /* cas rcursif */
  else {
    longueur p=n/2, q=n-p;                       /* demi-longueurs */
    chiffre *buff = getmem(2*q);
    int sa,sb;                                   /* signes des diffrences */
    chiffre r1,r2,r3;                            /* retenues des additions */

    sa = asub(c,   a+p, a, q,p);                 /* |a1-a0| */
    sb = asub(c+q, b+p, b, q,p);                 /* |b1-b0| */

    karatsuba(buff,  c,   c+q, q, n0);           /* |a1-a0||b1-b0| */
    karatsuba(c,     a,   b,   p, n0);           /* a0b0 */
    karatsuba(c+2*p, a+p, b+p, q, n0);           /* a1b1 */

    r1 = inc(     c+2*p,   c+p,   p,    p);      /* gamma + beta */
    r2 = add(c+p, c,       c+2*p, p,    p);      /* alpha + beta + gamma */
    r3 = inc(     c+2*p,   c+3*p, p,    2*q-p);  /* beta + gamma + delta */
         inc(     c+2*p,   &r1,   2*q,  1);      /* propage les retenues */
         inc(     c+2*p,   &r2,   2*q,  1);
         inc(     c+3*p,   &r1,   2*q-p,1);
         inc(     c+p+2*q, &r3,   p,    1);

    if (sa==sb) dec(c+p, buff, 2*q+p, 2*q);      /* retranche ...  */
    else        inc(c+p, buff, 2*q+p, 2*q);      /* (a1-a0)(b1-b0) */

    freemem(buff);
  }
}

/* test */
double test_karatsuba(longueur n, longueur n0, int nrep, int check) {
  chiffre *a, *b, *c;
  int i;
  double t;

  a = random_bits(n);
  b = random_bits(n);
  c = getmem(2*n);

  if (check) {
    chiffre *d = getmem(2*n);

    karatsuba(c,a,b,n,n0);
    muln2(d,a,b,n);
    if (compare(c,d,2*n)) {
      fprintf(stderr,"multiplication fausse\n");
      fprintf(stderr,"a="); dump(a,n);
      fprintf(stderr,"b="); dump(b,n);
      fprintf(stderr,"c="); dump(c,2*n);
      fprintf(stderr,"d="); dump(d,2*n);
      fflush(stderr);
      exit(1);
    }
    t = 0;
    freemem(d);
  }

  else {
    chrono();
    for (i=0; i < nrep; i++) karatsuba(c,a,b,n,n0);
    t = chrono()/nrep;
  }

  freemem(a); freemem(b); freemem(c);
  return(t);
}


/* multiplication par bloc */
void bloc_karatsuba(chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0) {

  if (n <= 1) bloc_muln2(c,a,b,1,n0);
  else {
    longueur p=n/2, q=n-p;                              /* demi-longueurs */
    chiffre *buff = getmem((2*q-1)*2*n0);

    bloc_sub(c,      a+p*n0, a, q, p, n0);              /* a1 - a0 */
    bloc_sub(c+q*n0, b+p*n0, b, q, p, n0);              /* b1 - b0 */

    bloc_karatsuba(buff,     c,      c+q*n0, q, n0);    /* (a1-a0)(b1-b0) */
    bloc_karatsuba(c,        a,      b,      p, n0);    /* a0b0 */
    bloc_karatsuba(c+4*p*n0, a+p*n0, b+p*n0, q, n0);    /* a1b1 */
    clearmem(c+4*p*n0-2*n0, 2*n0);                      /* complte a0b0 */

    bloc_inc(c+4*p*n0, c+2*p*n0,    p, p-1,     2*n0);  /* gamma + beta */
    bloc_add(c+2*p*n0, c, c+4*p*n0, p, p,       2*n0);  /* alpha + beta + gamma */
    bloc_inc(c+4*p*n0, c+6*p*n0,2*q-1, 2*q-p-1, 2*n0);  /* beta + gamma + delta */
    bloc_dec(c+2*p*n0, buff,    2*q+p, 2*q-1,   2*n0);  /* - (a1-a0)(b1-b0)     */

    freemem(buff);
  }
}

/* test */
double test_bloc_karatsuba(longueur n, longueur n0, int nrep, int check) {
  chiffre *a, *a1, *b, *b1, *c, *c1;
  longueur n1,p;
  int i;
  double t;

  for (n1=n; n1>=n0; n1>>=1); n1++;
  p = (n+n1-1)/n1;
  a = random_bits(n);  a1 = getmem(p*(n1+1)); expand(a1,a,n,p,n1);
  b = random_bits(n);  b1 = getmem(p*(n1+1)); expand(b1,b,n,p,n1);
  c = getmem(2*n);     c1 = getmem((2*p-1)*2*(n1+1));

  if (check) {
    chiffre *d = getmem(2*n);

    bloc_karatsuba(c1,a1,b1,p,n1+1);
    compact(c,c1,2*p-1,2*(n1+1),n1,2*n);
    muln2(d,a,b,n);
    if (compare(c,d,2*n)) {
      fprintf(stderr,"multiplication fausse\n");
      fprintf(stderr,"a ="); dump(a,n);
      fprintf(stderr,"a1="); bloc_dump(a1,p,n1+1);
      fprintf(stderr,"b ="); dump(b,n);
      fprintf(stderr,"b1="); bloc_dump(b1,p,n1+1);
      fprintf(stderr,"c ="); dump(c,2*n);
      fprintf(stderr,"c1="); bloc_dump(c1,2*p-1,2*(n1+1));
      fprintf(stderr,"d ="); dump(d,2*n);
      fflush(stderr);
      exit(1);
    }
    t = 0;
    freemem(d);
  }

  else {
    chrono();
    for (i=0; i < nrep; i++) {
      bloc_karatsuba(c1,a1,b1,p,n1+1);
      compact(c,c1,2*p-1,2*(n1+1),n1,2*n);
    }
    t = chrono()/nrep;
  }

  freemem(a); freemem(a1); freemem(b); freemem(b1); freemem(c); freemem(c1);
  return(t);
}

                    /* +----------------------+
                       |  carr de Karatsuba  |
                       +----------------------+ */

/* Version du produit adapte au cas o les deux facteurs sont gaux.   */
/* Pour dsactiver cette fonction et calculer un carr comme un produit */
/* ordinaire, dfinir  la compilation le symbole "nosquare".           */

/* c <- a^2, lg(a) = n */
void karasquare(chiffre *c, chiffre *a, longueur n, longueur n0) {

#ifdef nosquare
  karatsuba(c,a,a,n,n0);
#else

  if (n <= n0) sqrn2(c,a,n);
  else {
    longueur p=n/2, q=n-p;                       /* demi-longueurs */
    chiffre *buff = getmem(2*q);
    chiffre r1,r2,r3;                            /* retenues des additions */

    asub(c,   a+p, a, q,p);                      /* |a1-a0| */

    karasquare(buff,  c,   q, n0);               /* (a1-a0)^2 */
    karasquare(c,     a,   p, n0);               /* a0^2 */
    karasquare(c+2*p, a+p, q, n0);               /* a1^2 */

    r1 = inc(     c+2*p,   c+p,   p,    p);      /* gamma + beta */
    r2 = add(c+p, c,       c+2*p, p,    p);      /* alpha + beta + gamma */
    r3 = inc(     c+2*p,   c+3*p, p,    2*q-p);  /* beta + gamma + delta */
         inc(     c+2*p,   &r1,   2*q,  1);      /* propage les retenues */
         inc(     c+2*p,   &r2,   2*q,  1);
         inc(     c+3*p,   &r1,   2*q-p,1);
         inc(     c+p+2*q, &r3,   p,    1);

    dec(c+p, buff, 2*q+p, 2*q);                  /* retranche (a1-a0)^2 */

    freemem(buff);
  }
#endif
}

/* test */
double test_karasquare(longueur n, longueur n0, int nrep, int check) {
  chiffre *a, *c;
  int i;
  double t;

  a = random_bits(n);
  c = getmem(2*n);

  if (check) {
    chiffre *d = getmem(2*n);

    karasquare(c,a,n,n0);
    muln2(d,a,a,n);
    if (compare(c,d,2*n)) {
      fprintf(stderr,"carr faux\n");
      fprintf(stderr,"a="); dump(a,n);
      fprintf(stderr,"c="); dump(c,2*n);
      fprintf(stderr,"d="); dump(d,2*n);
      fflush(stderr);
      exit(1);
    }
    t = 0;
    freemem(d);
  }

  else {
    chrono();
    for (i=0; i < nrep; i++) karasquare(c,a,n,n0);
    t = chrono()/nrep;
  }

  freemem(a); freemem(c);
  return(t);
}

