                /* +------------------------------+
                   |  Produit tronqu de Mulders  |
                   +------------------------------+ */

#include "defs.h"

/* moiti basse */
void mulders_low(chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0) {
  longueur p = 0.6942363715*n, q=n-p;

  if ((n <= n0) || (q >= p)) muln2l(c,a,b,n);
  else {
    longueur i;
    chiffre *buff = getmem(2*p);
    ndouble r;

    /* produit central, retire les produits dpassant la diagonale */
    karatsuba(buff,a,b,p,n0);
    movemem(c,buff,n);
    r = buff[n] & mask;
    if (n+1 < 2*p) r += buff[n+1] << shift;
    for (i=q+1; i<p; i++) {
      r -= aa(i)*bb(n-i);
      if (i>q+1) r -= (aa(i)*bb(n+1-i)) << shift;
    }
    c[n] = r;
    c[n+1] = r >> shift;

    /* triangles latraux */
    if (q > 0){
      mulders_low(buff, a, b+p, q, n0); inc(c+p, buff, q+2, q+2);
      mulders_low(buff, a+p, b, q, n0); inc(c+p, buff, q+2, q+2);
    }

    freemem(buff);
  }
}

/* moiti haute */
void mulders_high(chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0) {
  longueur p = 0.6942363715*n, q=n-p;

  if ((n <= n0) || (q >= p)) muln2h(c,a,b,n);
  else {
    longueur i,j,k;
    chiffre *buff = getmem(2*p), t[3];
    ndouble r;

    /* produit central, retire les produits sous la diagonale */
    karatsuba(buff, a+q, b+q, p,n0);
    clearmem(t,3);
    for (i=q,j=n-q-2; j>=q; i++,j--) {
      r = aa(i)*bb(j) + (t[0] & mask);
      if (j > q) r += (aa(i)*bb(j-1)) >> shift;
      t[0] = r; r = (r >> shift) + (t[1] & mask);
      t[1] = r; r = (r >> shift) + (t[2] & mask);
      t[2] = r;
    }
    dec(buff+p-q-2,t,n+2,3);
    movemem(c,buff+p-q-1,n+1);

    /* triangles latraux */
    if (q > 0){
      mulders_high(buff, a, b+p, q, n0); inc(c, buff, n+1, q+1);
      mulders_high(buff, a+p, b, q, n0); inc(c, buff, n+1, q+1);
    }

    freemem(buff);
  }
}

/* moiti basse par blocs, c doit pouvoir contenir le produit complet */
void bloc_mulders_low(chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0) {
  longueur p = 0.6942363715*n, q=n-p;

  if (q >= p) bloc_muln2(c,a,b,n,n0);
  else {
    chiffre *buff = getmem((2*q-1)*2*n0);

    bloc_karatsuba(c,a,b,p,n0);
    bloc_mulders_low(buff, a,b+p*n0, q, n0); bloc_inc(c+2*p*n0, buff, q,q,2*n0);
    bloc_mulders_low(buff, b,a+p*n0, q, n0); bloc_inc(c+2*p*n0, buff, q,q,2*n0);

    freemem(buff);
  }
}

/* c <- ab mod base^(n+1), lg(a) = lg(b) = n  */
/* c doit pouvoir contenir le produit complet */
void raw_mulders_low(chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0) {
  longueur p = 0.6942363715*n, q=n-p;

  if (q >= p) muln2(c,a,b,n);
  else {
    chiffre *buff = getmem(2*q);
    ndouble r;

    karatsuba(c,a,b,p,n0);
    raw_mulders_low(buff, a,b+p, q, n0); inc(c+p, buff, q+1,q+1);
    raw_mulders_low(buff, b,a+p, q, n0); inc(c+p, buff, q+1,q+1);
    r = cc(n) + aa(p)*bb(q) + aa(q)*bb(p);
    c[n] = r;

    freemem(buff);
  }
}


/* test */
double test_mulders(int meth,longueur n, longueur n0, int nrep, int check) {
  chiffre *a, *b, *c;
  int i;
  double t;

  a = random_bits(n);
  b = random_bits(n);
  c = getmem(n+2);

  if (check) {
    chiffre *d = getmem(n+2);

    switch(meth) {
    case 1: mulders_low( c,a,b,n,n0); muln2l(d,a,b,n); break;
    case 2: mulders_high(c,a,b,n,n0); muln2h(d,a,b,n); c[n+1] = d[n+1] = 0; break;
    }
    if (compare(c,d,n+2)) {
      fprintf(stderr,"multiplication fausse\n");
      fprintf(stderr,"a="); dump(a,n);
      fprintf(stderr,"b="); dump(b,n);
      fprintf(stderr,"c="); dump(c,n+2);
      fprintf(stderr,"d="); dump(d,n+2);
      fflush(stderr);
      exit(1);
    }
    t = 0;
    freemem(d);
  }

  else {
    chrono();
    for (i=0; i < nrep; i++) switch(meth) {
    case 1: mulders_low( c,a,b,n,n0); break;
    case 2: mulders_high(c,a,b,n,n0); break;
    }
    t = chrono()/nrep;
  }

  freemem(a); freemem(b); freemem(c);
  return(t);
}
