                   /* +-------------------------+
                      |  Multiplication en n^2  |
                      +-------------------------+ */

#include "defs.h"

/* c <- ab, lg(a) = lg(b) = n */
void muln2(chiffre *c, chiffre *a, chiffre *b, longueur n) {
  longueur i,j;
  ndouble r;

  clearmem(c,n);
  for (i=0; i < n; i++,a++,c++) {
    for (j=0,r=0; j < n; j++) {
      r += aa(0)*bb(j) + cc(j);
      c[j] = r;
      r >>= shift;
    }
    c[n] = r;
  }
}

/* c <- partie basse(a*b), lg(a) = lg(b) = n */
void muln2l(chiffre *c, chiffre *a, chiffre *b, longueur n) {
  longueur i,j;
  ndouble r;

  clearmem(c,n+2);
  for (i=0; i < n; i++,a++,c++) {
    for (j=0,r=0; j < n-i; j++) {
      r += aa(0)*bb(j) + cc(j);
      c[j] = r;
      r >>= shift;
    }
    r += cc(n-i);
    c[n-i] = r;
    c[n-i+1] = cc(n-i+1) + (r >> shift);
  }
}

/* c <- partie haute(a*b), lg(a) = lg(b) = n */
void muln2h(chiffre *c, chiffre *a, chiffre *b, longueur n) {
  longueur i,j;
  ndouble r;

  c[0] = 0; c -= n-1; 
  for (i=0; i < n; i++,a++,c++) {
    for (j=n-i-1,r=0; j < n; j++) {
      r += aa(0)*bb(j) + cc(j);
      c[j] = r;
      r >>= shift;
    }
    c[n] = r;
  }
}

/* c <- partie mdianne(a*b), lg(a) = 2n-1, lg(b) = n */
void muln2m(chiffre *c, chiffre *a, chiffre *b, longueur n) {
  longueur i,j;
  ndouble r;

  clearmem(c,n+2); a += n-1;
  for (i=0; i<n; i++,a--) {
    for (j=0,r=0; j < n; j++) {
      r += aa(j)*bb(i) + cc(j);
      c[j] = r;
      r >>= shift;
    }
    r += cc(n);
    c[n] = r;
    c[n+1] = cc(n+1) + (r >> shift);
  }
}


/* multiplication par blocs en complment  2 */
void bloc_muln2(chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0) {

  longueur i,j;
  chiffre *buff;

  for (i=0; i < n; i++) {
    for (j=0; j < n; j++) {
      buff = c + ((i > 0) ? 2*(i+n-1)*n0 : 2*j*n0);
      muln2(buff, a+i*n0, b+j*n0, n0);
      if (aa(i*n0+n0-1) & msb) dec(buff+n0, b+j*n0, n0,n0);
      if (bb(j*n0+n0-1) & msb) dec(buff+n0, a+i*n0, n0,n0);
      if ((i > 0) && (j < n-1)) inc(c+2*(i+j)*n0, buff, 2*n0, 2*n0);
    }
  }
}

                       /* +----------------+
                          |  Carr en n^2  |
                          +----------------+ */

/* c <- a^2, lg(a) = n */
void sqrn2(chiffre *c, chiffre *a, longueur n) {
  longueur i,j;
  ndouble r;

  clearmem(c,2*n);

  /* doubles produits */
  for (i=1; i < n; i++) {
    for (j=0,r=0; j<i; j++) {
      r += aa(i)*aa(j) + cc(i+j);
      c[i+j] = r;
      r >>= shift;
    }
    c[2*i] = r;
  }
  for (i=0,r=0; i < 2*n; i++) {
    r += 2*cc(i);
    c[i] = r;
    r >>= shift;
  }

  /* carrs */
  for (i=0,r=0; i < n; i++) {
    r += aa(i)*aa(i) + cc(2*i);
    c[2*i] = r;
    r = (r >> shift) + cc(2*i+1);
    c[2*i+1] = r;
    r >>= shift;
  }
}


