#include "defs.h"
static chiffre un[] = {1};

                     /* +--------------------+
                        |  Division par Rmp  |
                        +--------------------+ */


/* c1 <- trunc(a/b1), a <- a - high(b1*c1),                     */
/* lg(a) = (2n+1)*n0, lg(b1) = n+1 blocs, lg(c1) = n blocs      */
/* b1 et c1 sont stocks sous forme de n ou n+1 blocs de taille */
/* n0+1, le chiffre de tte de chaque bloc tant nul.           */
/* a est stock sous forme compacte.                           */
/* en sortie, 0 <= a < b1.                                      */
/* le quotient peut tre surestim d'au plus 1.                 */
void bloc_rmpdiv(chiffre *c1, chiffre *a, chiffre *b1, longueur n, longueur n0) {
  longueur i;
  int cmp;

  /* compare a et b*base^n */
  for (i=n,cmp=0; (i>=0) && (cmp==0); i--)
    cmp = compare(a+(i+n)*n0, b1+i*(n0+1), n0);

  /* si a >= b*base^n tronque le quotient */
  if (cmp >= 0) {

    for (i=0; i<n; i++) c1[i] = mask;
    mix_dec(a+n*n0,b1,(n+1)*n0,n+1,n0+1,n0);
    for (i=0; i<n; i++) inc(a+(n-1)*n0, b1+i*(n0+1), (n+2)*n0, n0);
    inc(a+n*n0,b1+n*(n0+1),(n+1)*n0,n0);
  }

  /* cas de base : divise 3 blocs par 2 */
  else if (n <= 1) {
    chiffre *buff = getmem(2*n0);
    zdouble r,s;

    /* divise 2 blocs par 1 */
    div_knuth(c1, a+n0, b1+n0+1, n0);
    c1[n0] = 0;

    /* retranche c*low(b) */
    muln2(buff, c1,b1, n0);
    r = dec(a,buff,3*n0,2*n0);

    /* si ngatif, corrige */
    while (r) {dec(c1,un,n0,1); r += mix_inc(a,b1,3*n0,2,n0+1,n0);}

    freemem(buff);
  }

  /* cas rcursif */
  else {
    longueur p=n/2, q=n-p;
    chiffre *buff = getmem(2*q*(n0+1));
    zdouble r,s;

    /* divise les moitis hautes */
    bloc_rmpdiv(c1+q*(n0+1), a+2*q*n0, b1+q*(n0+1), p,n0);

    /* retranche rmp(b1,c1) */
    bloc_rmp4(q-p, buff, b1, c1+q*(n0+1), p, n0+1);
    r = mix_dec(a+(n-1)*n0, buff, (n+1)*n0, q, 2*n0+2, n0);

    /* si ngatif, corrige */
    while (r) {
      r += mix_inc(a+n*n0, b1+(n-q)*(n0+1), n*n0, q+1, n0+1, n0);
      for (i=q, s=1; (s); i++) {
	s = dec(c1+i*(n0+1),un,n0,1);
	r += inc(a+(n-1)*n0, b1+(n-i-1)*(n0+1), (n+1)*n0, n0);
      }
    }

    /* termine rcursivement */
    /* Pb : il se peut que a ait un chiffre de trop,  corriger */
    if (aa(p*n0+(2*q+1)*n0) != 0) {
      fprintf(stderr,"a trop grand dans rmpdiv\n");
      fflush(stderr);
      exit(1);
    }
    bloc_rmpdiv(c1, a+p*n0, b1+p*(n0+1), q,n0);

    freemem(buff);
  }
}

/* c <- trunc(a/b), a <- a mod b                   */
/* lg(a) = 2n, lg(b) = lg(c) = n                   */
/* le bit de poids fort de b doit tre non nul.    */
void rmpdiv(chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0) {
  chiffre *a1,*b1,*c1;
  longueur n1,n2,p;

  /* dcompose en blocs de sorte que le nb de blocs soit proche */
  /* d'une puissance de 2                                       */
  /* il faut dcaler b de sorte que son bit de de tte vale 1   */
  for (n1=n; n1>=n0; n1>>=1); n1++; p = (n+n1-1)/n1;
  n2 = p*n1 - n;

  a1 = getmem((2*p+1)*n1);
  b1 = getmem((p+1)*(n1+1));
  c1 = getmem(p*(n1+1));

  clearmem(a1,(2*p+1)*n1); movemem(a1+n1+n2,a,2*n);

  clearmem(b1,2*(n1+1));
  movemem(b1+n1+n2+1,b,n1-n2);
  expand(b1+2*(n1+1),b+n1-n2,n-n1+n2,p-1,n1);

  /* division sans reste */
  bloc_rmpdiv(c1,a1,b1,p,n1);
  compact(c,c1,p,n1+1,n1,n);

  /* calcule le reste */
  if (p>2) {
    chiffre *buff = getmem(2*n);

    raw_mulders_low(buff,b,c,n,n0);
    dec(a,buff,n+1,n+1); clearmem(a+n+1,n-1);
    while (aa(n)) {dec(c,un,n,1); inc(a,b,n+1,n);}

    freemem(buff);
  }
  else movemem(a,a1+n1+n2,2*n);

  freemem(a1); freemem(b1); freemem(c1);    
}

/* c <- trunc(a/b), a <- indtermin            */
/* lg(a) = 2n, lg(b) = lg(c) = n                */
/* le bit de poids fort de b doit tre non nul. */
void rmpdiv_nr(chiffre *c, chiffre *a, chiffre *b, longueur n, longueur n0) {
  chiffre *a1,*b1,*c1;
  longueur n1,n2,p;

  /* idem rmpdiv mais calcule avec n+1 chiffres */
  for (n1=n+1; n1>=n0; n1>>=1); n1++; p = n/n1 + 1;
  n2 = p*n1 - n;

  a1 = getmem((2*p+1)*n1);
  b1 = getmem((p+1)*(n1+1));
  c1 = getmem(p*(n1+1));

  clearmem(a1,(2*p+1)*n1); movemem(a1+n1+n2+1,a,2*n);

  clearmem(b1,2*(n1+1));
  movemem(b1+n1+n2+1,b,n1-n2);
  expand(b1+2*(n1+1),b+n1-n2,n-n1+n2,p-1,n1);

  /* division sans reste */
  bloc_rmpdiv(c1,a1,b1,p,n1);
  if (n1-1 < n) {
    compact(c+n1-1,c1+n1+1,p-1,n1+1,n1,n-n1+1);
    movemem(c,c1+1,n1-1);
  } else movemem(c,c1+1,n);

  /* si le premier chiffre aprs la virgule est trop petit calcule le reste */
  if ((p>2) && ((c1[0] & mask) == 0)) {
    chiffre *buff = getmem(2*n);

    raw_mulders_low(buff,b,c,n,n0);
    dec(a,buff,n+1,n+1); clearmem(a+n+1,n-1);
    while (aa(n)) {dec(c,un,n,1); inc(a,b,n+1,n);}

    freemem(buff);
  }

  freemem(a1); freemem(b1); freemem(c1);    
}

/* test */
double test_rmpdiv(int rem, longueur n, longueur n0, int nrep, int check) {
  chiffre *a,*b,*q,*r;
  double t;

  b = random_bits(n);   b[n-1] = bb(n-1) | msb;
  if (rem < 0) {
    q = random_bits(n);
    a = getmem(2*n);
    karatsuba(a,q,b,n,n0);
    rem = 0;
  }
  else {
    a = random_bits(2*n); a[2*n-1] = aa(2*n-1) % bb(n-1);
    q = getmem(n);
  }
  r = getmem(2*n);

  if (check) {
    chiffre *x = getmem(2*n);
    int ok;

    movemem(r,a,2*n);

    if (rem) {
      rmpdiv(q,r,b,n,n0);

      ok = (sub(x,b,r,n,n) == 0);
      if (ok) {
	muln2(x,q,b,n);
	inc(x,r,2*n,2*n); 
	ok = (compare(a,x,2*n)==0);
      }
    }

    else {
      rmpdiv_nr(q,r,b,n,n0);
      movemem(r,a,2*n);
      div_knuth(x,r,b,n); 
      ok = (compare(q,x,n) == 0);
    }

    if (!ok) {
      fprintf(stderr,"division fausse\n");
      fprintf(stderr,"a="); dump(a,2*n);
      fprintf(stderr,"b="); dump(b,n);
      fprintf(stderr,"q="); dump(q,n);
      fprintf(stderr,"r="); dump(r,2*n);
      fflush(stderr);
      exit(1);
    }
    
    t = 0;
    freemem(x);
  }

  else {
    int i;
    chrono();
    for (i=0; i < nrep; i++) {
      movemem(r,a,2*n);
      if (rem) rmpdiv(q,r,b,n,n0); else rmpdiv_nr(q,r,b,n,n0);
    }
    t = chrono()/nrep;
  }

  freemem(a); freemem(b); freemem(q); freemem(r);
  return(t);
}





