#include "defs.h"
static chiffre un[] = {1};

               /* +--------------------------------+
                  |  Racine carre traditionnelle  |
                  +--------------------------------+ */

/* c <- 2*trunc(a^1/2), a <- a - c^2/4, lg(a) = 2n */
/* le chiffre de poids fort de a doit tre compris */
/* entre base/16 inclus et base/4 exclus           */

void sqrtn2(chiffre *c, chiffre *a, longueur n) {
  longueur i,j;
  ndouble q,r,x,y;
  zdouble z;

  /* racine carre des deux chiffres de tte */
  r = aa(2*n-2) + (aa(2*n-1) << shift);
  for (x=mask/2,y=x+1; x < y;) {y = x; x = (x + r/x)/2;}
  r -= y*y; a[2*n-2] = r; a[2*n-1] = r >> shift;
  y <<= 1;  c[n-1] = y;

  /* calcule les chiffres suivants par division */
  for (i=n-2; i>=0; i--) {

    if (aa(n+i+1)) q = mask;
    else {
      x = (aa(i+n) << shift) + aa(i+n-1);
      q = x/y;
      if (q > mask) q = mask;
    }
    c[i] = q;

    /* a <- a - c*q - q^2 */
    for (j=i,r=0; j <= n+1; j++) {
      if (j < n) r += q*cc(j);
      z = aa(i+j) - (r&mask);
      a[i+j] = z;
      r >>= shift;
      if (z < 0) r++;
    }

    /* si ngatif corrige */
    while(z) {
      z += inc(a+2*i, c+i, n-i+2, n-i);
      q--; c[i] = q;
      z += inc(a+2*i, c+i, n-i+2, 1);
    }

    /* double le chiffre ajout */
    for (j=i,r=q; (r); j++) {
      r += cc(j);
      c[j] = r;
      r >>= shift;
    }
  }
}


                  /* +---------------------------+
                     |  Racine carre rcursive  |
                     +---------------------------+ */


/* c <- 2*trunc(a^1/2), lg(a) = 2n, lg(c) = n                       */
/* si rem = 0 alors a <- a - high(c^2/4) sinon a <- a - c^2/4       */
/* si rem = 0 le quotient est trop grand d'au plus 2+2log(n) units */
/* le chiffre de poids fort de a doit tre compris entre            */
/* base/16 inclus et base/4 exclu                                   */

void sqrtrec(int rem, chiffre *c, chiffre *a, longueur n, longueur n0) {

  if (n <= n0) sqrtn2(c,a,n);
  else {
    longueur p=n/2, q=n-p;
    chiffre *buff = getmem(n);
    zdouble z;

    /* racine carre de la partie haute */
    sqrtrec(1,c+p, a+2*p,q,n0);

    /* partie basse par division tronque */
    if (compare(a+n,c+q,p)< 0) burnikel(rem,c,a+q,c+q,p,n0);
    else {
      longueur i,j;

      for (i=0; i<p; i++) c[i] = mask;
      clearmem(a+n,p);
      for (i=p; i>0;) {
	j = ((i <= n0) || (rem)) ? i : (i+1)/2;
	inc(a+n-j,c+n-i,j+1,j);
	i -= j;
      }
    }

    if (rem) {

      /* rectangle q x p -> carr + bande ventuelle */
      karasquare(buff,c,p,n0);

      if (q > p) {
	longueur i;
	ndouble r;

	for (i=0, r=0; i < p; i++) {
	  r += cc(p)*cc(i) + (buff[i+p] & mask);
	  buff[i+p] = r;
	  r >>= shift;
	}
	buff[n-1] = r;
      }

      /* double la partie basse et retire le rectangle */
      inc(c,c,p+1,p);
      z = dec(a,buff,n+p+1,n);

      while (z) {dec(c,un,p+1,1); z += inc(a,c,n+p+1,n); dec(c,un,p+1,1);}
    }

    freemem(buff);
  }
}
    
/* corrige la racine sans reste */
void sqrt_correct(chiffre *c, chiffre *a, longueur n, longueur n0) {

  if (n > n0) {
    longueur p=n/2, q=n-p;
    chiffre *buff = getmem(n);
    zdouble z;

    /* corrige le reste de la deuxime division */
    burnikel_correct(c,a+q,c+q,p,n0);

    /* double la partie basse et retire le rectangle */
    karasquare(buff,c,p,n0);

    if (q > p) {
      longueur i;
      ndouble r;
      
      for (i=0, r=0; i < p; i++) {
	r += cc(p)*cc(i) + (buff[i+p] & mask);
	buff[i+p] = r;
	r >>= shift;
      }
      buff[n-1] = r;
    }
    
    inc(c,c,p+1,p);
    z = dec(a,buff,n+p+1,n);
    
    while (z) {dec(c,un,p+1,1); z += inc(a,c,n+p+1,n); dec(c,un,p+1,1);}
    freemem(buff);
  }
}

                    /* +----------------------------+
                       |  Racine carre sans reste  |
                       +----------------------------+ */

void sqrt_nr(chiffre *c, chiffre *a, longueur n, longueur n0) {

  if (n <= n0) sqrtn2(c,a,n);
  else {

    /* calcule la racine sans reste avec deux chiffres aprs la virgule */
    chiffre *a1 = getmem(2*n+4);
    chiffre *c1 = getmem(n+2);

    movemem(a1+4,a,2*n); clearmem(a1,4);
    sqrtrec(0,c1,a1,n+2,n0);

    /* si la partie dcimale est trop petite corrige la division */
    if ((c1[1] & mask) == 0) sqrt_correct(c1,a1,n+2,n0);
    else inc(c1,c1,n/2+2,n/2+1);

    /* tronque le rsultat */
    movemem(c,c1+2,n); c[0] = c[0] & (mask-1);

    freemem(a1); freemem(c1);
  }
}

/* division par 2 */
static void div2(chiffre *a, longueur n) {
  longueur i;
  ndouble r;

  for (i=n-1,r=0; i>=0; i--) {
    r += aa(i);
    if (r%2) {a[i] = r/2; r = mask+1;} else {a[i] = r/2; r = 0;}
  }
}

/* test */
double test_sqrt(int rem, longueur n, longueur n0, int nrep, int check) {
  chiffre *a,*q,*r;
  double t;

  if (rem < 0) {
    q = random_bits(n); q[n-1] = (q[n-1] & (mask/2)) | (msb/2);
    a = getmem(2*n);
    karasquare(a,q,n,n0);
    rem = 0;
  }
  else {
    a = random_bits(2*n); a[2*n-1] = (aa(2*n-1) & (mask/4)) | (msb/4);
    q = getmem(n);
  }
  r = getmem(2*n);

  if (check) {
    chiffre *x = getmem(2*n);
    int ok;

    movemem(r,a,2*n);

    if (rem) {
      sqrtrec(1,q,r,n,n0);
      ok = (sub(x,q,r,n,n) == 0);
      if (ok) {
	div2(q,n);
	sqrn2(x,q,n);
	inc(x,r,2*n,2*n); 
	ok = (compare(a,x,2*n) == 0);
      }
    }

    else {
      sqrt_nr(q,r,n,n0);
      movemem(r,a,2*n);
      sqrtn2(x,r,n);
      ok = (compare(q,x,n) == 0);
      /*      div2(q,n); */
    }

    if (!ok) {
      fprintf(stderr,"racine fausse\n");
      fprintf(stderr,"a="); dump(a,2*n);
      fprintf(stderr,"q="); dump(q,n);
      fprintf(stderr,"r="); dump(r,2*n);
      fflush(stderr);
      exit(1);
    }
    
    t = 0;
    freemem(x);
  }

  else {
    int i;
    chrono();
    for (i=0; i < nrep; i++) {
      movemem(r,a,2*n);
      if (rem) sqrtrec(1,q,r,n,n0); else sqrt_nr(q,r,n,n0);
      div2(q,n);
    }
    t = chrono()/nrep;
  }

  freemem(a); freemem(q); freemem(r);
  return(t);
}

